//===================================================================
//
//  InfoNES_System_GP32.cpp : GP32 specific file
//
//  2003/01/08  GP32 port started [ Anders Granlund ]
//  2001/05/18  InfoNES Project ( Sound is based on DarcNES )
//
//===================================================================

//-------------------------------------------------------------------
//  Include files
//-------------------------------------------------------------------
#include "gp32.h"
#include "enablecache.h"
#include "debug.h"
#include "./../InfoNES.h"
#include "./../InfoNES_System.h"
#include "./../InfoNES_pAPU.h"
#include "./../K6502.h"


//-------------------------------------------------------------------
//  Constants and defines
//-------------------------------------------------------------------

#define USE_OVERCLOCK		1
#define USE_SOUND 		1
#define DEFAULT_FRAMESKIP	2

#define APPLICATION_TITLE	"InfoNES-GP32 v0.76J"
#define	KEYMASK_A		(1<<0)
#define	KEYMASK_B		(1<<1)
#define	KEYMASK_SELECT		(1<<2)
#define	KEYMASK_START		(1<<3)
#define	KEYMASK_UP		(1<<4)
#define	KEYMASK_DOWN		(1<<5)
#define	KEYMASK_LEFT		(1<<6)
#define	KEYMASK_RIGHT		(1<<7)

#define HARDWARE_PALETTE	((dword *)0x14A00400)

//-------------------------------------------------------------------
//  ROM image file information
//-------------------------------------------------------------------
char	szRomName[ 256 ];
char	szSaveName[ 256 ];
int	nSRAM_SaveFlag;


//-------------------------------------------------------------------
//  Global Variables ( GP32 specific )
//-------------------------------------------------------------------

GP_PALETTEENTRY	HwPalette[256];
GP_PALETTEENTRY	gpPalette[256];
GP_HPALETTE	gpPalCur;
GP_HPALETTE	gpPalOld;
GPDRAWSURFACE	gpScreen[2];
int		iscreen;
int		ticker;
volatile long		nes_timer;
volatile long		gp_timer;


//-------------------------------------------------------------------
//  Function prototypes ( GP32 specific )
//-------------------------------------------------------------------
char*	FileSelector(char *path);	// File selector
int	LoadSRAM();					// Load SRAM from SMC
int	SaveSRAM();					// Save SRAM to SMC


//-------------------------------------------------------------------
// NES palette data
//-------------------------------------------------------------------
WORD NesPalette[ 64 ] =
{
  0x39ce, 0x1071, 0x0015, 0x2013, 0x440e, 0x5402, 0x5000, 0x3c20,
  0x20a0, 0x0100, 0x0140, 0x00e2, 0x0ceb, 0x0000, 0x0000, 0x0000,
  0x5ef7, 0x01dd, 0x10fd, 0x401e, 0x5c17, 0x700b, 0x6ca0, 0x6521,
  0x45c0, 0x0240, 0x02a0, 0x0247, 0x0211, 0x0000, 0x0000, 0x0000,
  0x7fff, 0x1eff, 0x2e5f, 0x223f, 0x79ff, 0x7dd6, 0x7dcc, 0x7e67,
  0x7ae7, 0x4342, 0x2769, 0x2ff3, 0x03bb, 0x0000, 0x0000, 0x0000,
  0x7fff, 0x579f, 0x635f, 0x6b3f, 0x7f1f, 0x7f1b, 0x7ef6, 0x7f75,
  0x7f94, 0x73f4, 0x57d7, 0x5bf9, 0x4ffe, 0x0000, 0x0000, 0x0000
};


#if(ENABLE_DEBUG>0)
void GpDebugText(char *str){
	int key;
	Palette_GP32();
	GpFillScreen(0xe0);
	GpTextOut(NULL, &gpScreen[iscreen], 10, 10, str, 0);

	GpSwapScreen();

	GpKeyGetEx(&key);while(key!=0){ GpKeyGetEx(&key); }
	GpKeyGetEx(&key);while(key==0){ GpKeyGetEx(&key); }

}
#endif


//-------------------------------------------------------------------
// Palette_NES
//-------------------------------------------------------------------
void Palette_NES(){
	int i;
	// Setup NES palette
	for (i = 0; i < 64; i++) {
		gpPalette[i] = (NesPalette[i] << 1) | 1;
	}
	for (i = 0; i < 64; i++) {
		gpPalette[i+64] = (NesPalette[i] << 1) | 1;
	}

	gpPalette[254] = 0xFFFE;
	gpPalette[255] = 0;

	gpPalCur = GpPaletteCreate(256, gpPalette);
	GpPaletteSelect(gpPalCur);
	GpPaletteRealize();
}

//-------------------------------------------------------------------
// Palette_GP32
//-------------------------------------------------------------------
void Palette_GP32(){
	int i;
	// Setup GP32 palette
	for (i = 0; i < 256; i++) {
		gpPalette[i] = HwPalette[i];
	}
	gpPalCur = GpPaletteCreate(256, gpPalette);
	GpPaletteSelect(gpPalCur);
	GpPaletteRealize();
}

//-------------------------------------------------------------------
// InfoNES_GetScreenPtr
//-------------------------------------------------------------------
unsigned char *InfoNES_GetScreenPtr(){
	return gpScreen[iscreen].ptbuffer;
}


//-------------------------------------------------------------------
// GpTimer
//-------------------------------------------------------------------
void GpTimer(void) {
	gp_timer++;

}


//-------------------------------------------------------------------
// GpFillScreen
//-------------------------------------------------------------------
void GpFillScreen(unsigned char col){
	gm_memset( gpScreen[iscreen].ptbuffer, col, 320*240);
}

//-------------------------------------------------------------------
// GpSwapScreen
//-------------------------------------------------------------------
void GpSwapScreen(){
	GpSurfaceSet( &gpScreen[iscreen] );
	iscreen = 1 - iscreen;
}

//-----------------------------------------------------------------------------
// Sleep() wrapper.
//
//	Wait for a designated number of milliseconds.
//
//-----------------------------------------------------------------------------
void Sleep(unsigned int ms){
	unsigned int n_tick;
	n_tick = GpTickCountGet();
	while ( ( GpTickCountGet() - n_tick ) < ms)
	;

}

//===================================================================
//
//            InfoNES_Wait() : Wait Emulation if required
//
//===================================================================
void InfoNES_Wait(){
#if(DEBUG_NES>0)
        char str[1024];
	int key;
	BYTE a;
	char c;
	int b;

	a = 0xFB;
	c = (signed char)a;
	b = a;
	if(a>127) b = -(256 - a);

        gm_memset( gpScreen[iscreen].o_buffer, 0xe0, 320*240);
        gm_sprintf(str, "PC = %4x, SP = %4x\n", PC-1, SP);
	GpTextOut(NULL, &gpScreen[iscreen], 10, 10, str, 0);

        gm_sprintf(str, "FAXY = %2x,%2x,%2x,%2x\n", F, A, X, Y);
 	GpTextOut(NULL, &gpScreen[iscreen], 10, 30, str, 0);

        gm_sprintf(str, "CODE = %2x, DBG = %2x, iDBG = %i\n", OPCODE, 0/*DEBUGNUM*/, c);
	GpTextOut(NULL, &gpScreen[iscreen], 10, 50, str, 0);
	GpSwapScreen();

	Sleep(100);

	GpKeyGetEx(&key);
	while(key!=0){
		GpKeyGetEx(&key);
	}
	while(key!=GPC_VK_FA){
		GpKeyGetEx(&key);
	}

#endif
}

//===================================================================
//
//                main() : Application main
//
//===================================================================
void GpMain(void *arg)
{
	int i;
	char *vid;
	char str[1024];


	//-------------------------------------------------------------------
	//  Initialize file system
	//-------------------------------------------------------------------
	GpFatInit();

	//-------------------------------------------------------------------
	// Palette
	//-------------------------------------------------------------------
	for (i = 0; i < 256; i++)
		HwPalette[i] = HARDWARE_PALETTE[i];

	//-------------------------------------------------------------------
	//  Main Loop
	//-------------------------------------------------------------------
	while(1){

		EnableCache();

		#if(USE_OVERCLOCK>0)
		// Back to normal speed during file selector
		GpClockSpeedChange (40000000, 0x48013, 1);
		#endif

		Palette_GP32();

		//-------------------------------------------------------------------
		//  Do file selector
		//-------------------------------------------------------------------
		gm_strcpy(szRomName, (char*)fileselector("gp:\\gpmm\\nes\\", 0));
		//gm_sprintf(szRomName, "gp:\\gpmm\\gunsmoke.nes");

		// Load ROM image
		if(InfoNES_Load(szRomName) == 0){

			#if(USE_OVERCLOCK>0)
 			GpClockSpeedChange (132000000, 0x3a011, 2);
			#endif

			//-------------------------------------------------------------------
			//  Initialize Video
			//-------------------------------------------------------------------
			GpGraphicModeSet(8, NULL);		// We want 8bit video
			GpLcdSurfaceGet( &gpScreen[0], 0);		// Get screen
			GpLcdSurfaceGet( &gpScreen[1], 1);		// Get screen
			GpSurfaceSet( &gpScreen[0] );		// Set the front buffer
			iscreen = 1;
			if( !(GPC_LCD_ON_BIT & GpLcdStatusGet()))
				GpLcdEnable();

			// Clear both buffers
			GpFillScreen(255);
			GpSwapScreen();
			GpFillScreen(255);
			GpSwapScreen();

			Palette_NES();


			//-------------------------------------------------------------------
			// Initialize core
			//-------------------------------------------------------------------
			nes_timer = 0;
			gp_timer = 0;
			GpTimerOptSet(0, 60, 0, GpTimer);
			GpTimerSet(0);

			LoadSRAM();
			InfoNES_Init();
			InfoNES_SetFrameSkip(DEFAULT_FRAMESKIP);
			InfoNES_SetRenderType(RENDERER_LINE);

			// Emulate
			InfoNES_Cycle();

			// Save SRAM to file
			SaveSRAM();

			// Shut down the core
			InfoNES_Fin();

			// Shut down GP stuff
			GpTimerKill(0);
			GpPcmStop();

		}
	}
}



//===================================================================
//
//           LoadSRAM() : Load a SRAM
//
//===================================================================
int LoadSRAM()
{

  F_HANDLE fp;
  unsigned char pSrcBuf[ SRAM_SIZE ];
  unsigned char chData;
  unsigned char chTag;
  int nRunLen;
  int nDecoded;
  int nDecLen;
  int nIdx;
  int i,j;
  int readcount;

  // It doesn't need to save it
  nSRAM_SaveFlag = 0;

  // It is finished if the ROM doesn't have SRAM
  if ( !ROM_SRAM )
    return 0;

  // There is necessity to save it
  nSRAM_SaveFlag = 1;

  GpDirCreate("gp:\\gpmm\\nes\\sav", NOT_IF_EXIST);
  GpRelativePathSet("gp:\\gpmm\\nes\\sav\\");

  // The preparation of the SRAM file name
  gm_strcpy( szSaveName, szRomName );
  j = -1;
  i = 0;
  while(i<gm_lstrlen(szSaveName) && j==-1){
  	if( szSaveName[i] == '.')
  		j = i+1;
  	i++;
  }
  for(i=0;i<strlen(szSaveName);i++)
  gm_strcpy( &szSaveName[j], "srm" );

  //-------------------------------------------------------------------
  //  Read a SRAM data
  //-------------------------------------------------------------------

  // Open SRAM file
//  fp = fopen( szSaveName, "rb" );
  if(GpFileGetSize(szSaveName, &readcount) != SM_OK)
    return -1;

  if(GpFileOpen(szSaveName, OPEN_R, &fp) != SM_OK)
    return -1;

  // Read SRAM data
  //fread( pSrcBuf, SRAM_SIZE, 1, fp );
  GpFileRead(fp, pSrcBuf, SRAM_SIZE, &readcount);

  // Close SRAM file
  //fclose( fp );
  GpFileClose(fp);

  //-------------------------------------------------------------------
  //  Extract a SRAM data
  //-------------------------------------------------------------------

  nDecoded = 0;
  nDecLen = 0;

  chTag = pSrcBuf[ nDecoded++ ];

  while ( nDecLen < 8192 )
  {
    chData = pSrcBuf[ nDecoded++ ];

    if ( chData == chTag )
    {
      chData = pSrcBuf[ nDecoded++ ];
      nRunLen = pSrcBuf[ nDecoded++ ];
      for ( nIdx = 0; nIdx < nRunLen + 1; ++nIdx )
      {
        SRAM[ nDecLen++ ] = chData;
      }
    }
    else
    {
      SRAM[ nDecLen++ ] = chData;
    }
  }

  // Successful
  return 0;
}

//===================================================================
//
//           SaveSRAM() : Save a SRAM
//
//===================================================================
int SaveSRAM()
{
//  FILE *fp;
  F_HANDLE fp;
  int nUsedTable[ 256 ];
  unsigned char chData;
  unsigned char chPrevData;
  unsigned char chTag;
  int nIdx;
  int nEncoded;
  int nEncLen;
  int nRunLen;
  unsigned char pDstBuf[ SRAM_SIZE ];
  int readcount;

  if ( !nSRAM_SaveFlag )
    return 0;  // It doesn't need to save it

  //-------------------------------------------------------------------
  //  Compress a SRAM data
  //-------------------------------------------------------------------

  gm_memset( nUsedTable, 0, sizeof nUsedTable );

  for ( nIdx = 0; nIdx < SRAM_SIZE; ++nIdx )
  {
    ++nUsedTable[ SRAM[ nIdx++ ] ];
  }
  for ( nIdx = 1, chTag = 0; nIdx < 256; ++nIdx )
  {
    if ( nUsedTable[ nIdx ] < nUsedTable[ chTag ] )
      chTag = nIdx;
  }

  nEncoded = 0;
  nEncLen = 0;
  nRunLen = 1;

  pDstBuf[ nEncLen++ ] = chTag;

  chPrevData = SRAM[ nEncoded++ ];

  while ( nEncoded < SRAM_SIZE && nEncLen < SRAM_SIZE - 133 )
  {
    chData = SRAM[ nEncoded++ ];

    if ( chPrevData == chData && nRunLen < 256 )
      ++nRunLen;
    else
    {
      if ( nRunLen >= 4 || chPrevData == chTag )
      {
        pDstBuf[ nEncLen++ ] = chTag;
        pDstBuf[ nEncLen++ ] = chPrevData;
        pDstBuf[ nEncLen++ ] = nRunLen - 1;
      }
      else
      {
        for ( nIdx = 0; nIdx < nRunLen; ++nIdx )
          pDstBuf[ nEncLen++ ] = chPrevData;
      }

      chPrevData = chData;
      nRunLen = 1;
    }

  }
  if ( nRunLen >= 4 || chPrevData == chTag )
  {
    pDstBuf[ nEncLen++ ] = chTag;
    pDstBuf[ nEncLen++ ] = chPrevData;
    pDstBuf[ nEncLen++ ] = nRunLen - 1;
  }
  else
  {
    for ( nIdx = 0; nIdx < nRunLen; ++nIdx )
      pDstBuf[ nEncLen++ ] = chPrevData;
  }

  //-------------------------------------------------------------------
  //  Write a SRAM data
  //-------------------------------------------------------------------

  GpDirCreate("gp:\\gpmm\\nes\\sav", NOT_IF_EXIST);
  GpRelativePathSet("gp:\\gpmm\\nes\\sav\\");

  // Open SRAM file
  //fp = fopen( szSaveName, "wb" );
  if(GpFileCreate(szSaveName, ALWAYS_CREATE, &fp) != SM_OK)
  	return -1;

//  if ( fp == NULL )
//    return -1;

  // Write SRAM data
  //fwrite( pDstBuf, nEncLen, 1, fp );
  GpFileWrite(fp, pDstBuf, nEncLen);

  // Close SRAM file
  //fclose( fp );
  GpFileClose(fp);

  // Successful
  return 0;
}


//===================================================================
//
//               InfoNES_ReadRom() : Read ROM image file
//
//===================================================================
int InfoNES_ReadRom( char *pszFileName )
{
	F_HANDLE fp = NULL;
	unsigned long readcount;


	// Test if file exists
	if(GpFileGetSize(pszFileName, &readcount) != SM_OK){
		return -1;
	}

	// Open rom image
	if(GpFileOpen(pszFileName, OPEN_R, &fp) != SM_OK)
		return -1;

	// Read ROM header
	GpFileRead(fp, &NesHeader, sizeof(NesHeader), &readcount);

	// Validate
	if(NesHeader.byID[0] != 'N' || NesHeader.byID[1] != 'E' || NesHeader.byID[2] != 'S'){
		GpFileClose( fp);
		return -1;
	}


	// Clear SRAM
	gm_memset(SRAM, 0, SRAM_SIZE);

	// If trainer presents Read Triner at 0x7000-0x71ff
	if ( NesHeader.byInfo1 & 4 )
	{
		GpFileRead( fp, &SRAM[ 0x1000 ], 512, &readcount );
	}

	// Allocate Memory for ROM Image
	ROM = (BYTE *)gm_malloc( NesHeader.byRomSize * 0x4000 );

	// Read ROM Image
	GpFileRead( fp, ROM, 0x4000 * NesHeader.byRomSize, &readcount );

	if ( NesHeader.byVRomSize > 0 )
	{
		// Allocate Memory for VROM Image
		VROM = (BYTE *)gm_malloc( NesHeader.byVRomSize * 0x2000 );

		// Read VROM Image
		GpFileRead( fp, VROM, 0x2000 * NesHeader.byVRomSize, &readcount );
	}

	// File close
	GpFileClose( fp );

	// Successful
	return 0;
}

/*===================================================================*/
/*                                                                   */
/*           InfoNES_ReleaseRom() : Release a memory for ROM         */
/*                                                                   */
/*===================================================================*/
void InfoNES_ReleaseRom()
{

	if ( ROM )
	{
		gm_free( ROM );
		ROM = NULL;
	}

	if ( VROM )
	{
		gm_free( VROM );
		VROM = NULL;
	}
}

//===================================================================
//
//             InfoNES_MemoryCopy() : memcpy
//
//===================================================================
/*
void *InfoNES_MemoryCopy( void *dest, const void *src, int count )
{
  gm_memcpy( dest, src, count );
  return dest;
}

//===================================================================
//
//             InfoNES_MemorySet() : memset
//
//===================================================================
void *InfoNES_MemorySet( void *dest, int c, int count )
{
  gm_memset( dest, c, count);
  return dest;
}
*/


//===================================================================
//
//             InfoNES_PadState() : Get a joypad state
//
//===================================================================
void InfoNES_PadState( DWORD *pdwPad1, DWORD *pdwPad2, DWORD *pdwSystem )
{
	int key;
	DWORD pad;

	//-----------------------------------------------------------
	// JOYSTICK 1
	//-----------------------------------------------------------
	GpKeyGetEx(&key);

	pad = 0;
	if(key & GPC_VK_UP)	pad |= KEYMASK_UP;
	if(key & GPC_VK_DOWN)	pad |= KEYMASK_DOWN;
	if(key & GPC_VK_LEFT)	pad |= KEYMASK_LEFT;
	if(key & GPC_VK_RIGHT)	pad |= KEYMASK_RIGHT;
	if(key & GPC_VK_FA)	pad |= KEYMASK_A;
	if(key & GPC_VK_FB)	pad |= KEYMASK_B;
	if(key & GPC_VK_START)	pad |= KEYMASK_START;
	if(key & GPC_VK_SELECT)	pad |= KEYMASK_SELECT;
	*pdwPad1 = pad;

	//-----------------------------------------------------------
	// JOYSTICK 2
	//-----------------------------------------------------------
	*pdwPad2 = 0;

	//-----------------------------------------------------------
	// SYSTEM KEYS
	//-----------------------------------------------------------
	*pdwSystem = 0;
	if(key & GPC_VK_FL && key & GPC_VK_FR)
		*pdwSystem = PAD_SYS_QUIT;

//  *pdwPad1   = dwKeyPad1;
//  *pdwPad2   = dwKeyPad2;
//  *pdwSystem = dwKeySystem;
}



#if(pAPU_QUALITY>1)
#define BUFLEN (735)
#else
#define BUFLEN (368+1)
#endif
#define BUFCOUNT 4
unsigned char *sndbuffer;
unsigned char *sndbuffer_pos;
unsigned char *sndbuffer_end;
int      	*play_pos;
volatile int	play_chan;

//===================================================================
//
//        InfoNES_SoundInit() : Sound Emulation Initialize
//
//===================================================================
void InfoNES_SoundInit( void )
{
#if(USE_SOUND>0)

	sndbuffer = gm_malloc(BUFLEN*BUFCOUNT);

	sndbuffer_pos = sndbuffer + (BUFLEN*BUFCOUNT/2);
	sndbuffer_end = sndbuffer + (BUFLEN*BUFCOUNT);

#if(pAPU_QUALITY>1)
	GpPcmInit(PCM_M44,PCM_8BIT);
#else
	GpPcmInit(PCM_M22,PCM_8BIT);
#endif
	gm_memset(sndbuffer, 0, BUFLEN*BUFCOUNT);
	GpPcmPlay(sndbuffer, BUFLEN*BUFCOUNT, 1);
	GpPcmLock((unsigned char *)sndbuffer, (int*)&play_chan, (unsigned int *)&play_pos);
#endif
}

//===================================================================
//
//        InfoNES_SoundOpen() : Sound Open
//
//===================================================================
int InfoNES_SoundOpen( int samples_per_sync, int sample_rate )
{
#if(USE_SOUND>0)
	return 1;
#else
	return 0;
#endif
}

//===================================================================
//
//        InfoNES_SoundClose() : Sound Close
//
//===================================================================
void InfoNES_SoundClose( void )
{
#if(USE_SOUND>0)
	GpPcmStop();
	gm_free(sndbuffer);
#endif
}

//===================================================================
//
//            InfoNES_SoundOutput4() : Sound Output 4 Waves
//
//===================================================================
void InfoNES_SoundOutput4( int samples, BYTE *wave1, BYTE *wave2, BYTE *wave3, BYTE *wave4 )
{
#if(USE_SOUND>0)

	int i;
	unsigned int pp,pw,dist;

	//play_pos = 0;

	pp=((unsigned int)*play_pos);
	pw=((unsigned int)sndbuffer_pos);
	if (pw>pp) dist=(pw-pp);
	else dist=((pw+(BUFLEN*BUFCOUNT))-pp);
	if (dist<800) gp_timer++;  //push a frame rendering to resync the audio!!


	for (i = BUFLEN; i !=0; i--)
		*(sndbuffer_pos++) = ((*(wave1++) + *(wave2++) + *(wave3++) + *(wave4)) >> 2);

	if(sndbuffer_pos >= sndbuffer_end)	sndbuffer_pos = sndbuffer;


	// Play the sound
//	GpPcmPlay((unsigned char*)sndbuffer, BUFLEN, 0);
#endif
}

// wave2 = pickups in smb1


//===================================================================
//
//            InfoNES_Menu() :
//
//===================================================================
int InfoNES_Menu(){
	return 0;
}








/*===================================================================*/
/*                                                                   */
/*      InfoNES_LoadFrame() :                                        */
/*           Transfer the contents of work frame on the screen       */
/*                                                                   */
/*===================================================================*/

void InfoNES_LoadFrame(int show)
{
	register int x asm("r8");
	register int y asm("r7");
	register BYTE *pScreen asm("r5");
	register BYTE *pBuffer asm("r6");
	nes_timer++;

	if (((int)(gp_timer-nes_timer))>3) gp_timer=nes_timer+1; // max 3 frames lost
	while(gp_timer < nes_timer){
		asm("nop");
	}

	if(show){
		pScreen = (BYTE *) gpScreen[iscreen].ptbuffer + 239 + (32*240);
		pBuffer = (BYTE *) WorkFrame;

		for(y=NES_DISP_HEIGHT; y!=0; y--){
			for(x=NES_DISP_WIDTH; x!=0; x--){
				*(pScreen) = *(pBuffer++); pScreen += 240;
			}
			pScreen -= ((NES_DISP_WIDTH*240) + 1);
		}
		GpSurfaceSet( &gpScreen[iscreen] );
		iscreen = 1 - iscreen;
	}
}

